from labelbox import Client as labelboxClient
from labelbox import Dataset as labelboxDataset
from labelbox import Project as labelboxProject
import uuid
import math

def create_global_key_to_data_row_dict(client:labelboxClient, global_keys:list):
    """ Creates a dictionary where {key=global_key : value=data_row_id}
    Args:
        client          :   Required (labelbox.client.Client) - Labelbox Client object    
        global_keys     :   Required (list(str)) - List of global key strings
    Returns:
        Dictionary where {key=global_key : value=data_row_id}
    """
    res = client.get_data_row_ids_for_global_keys(global_keys, timeout_seconds=240)
    if res['errors']:
        raise ValueError(f"{res}")
    global_key_to_data_row_dict = {global_keys[i] : res['results'][i] for i in range(0, len(global_keys))}
    return global_key_to_data_row_dict

def check_global_keys(client:labelboxClient, global_keys:list):
    """ Checks if data rows exist for a set of global keys
    Args:
        client                  :   Required (labelbox.client.Client) - Labelbox Client object    
        global_keys             :   Required (list(str)) - List of global key strings
    Returns:
        GQL payload             :   Dictionary where {keys="accessDeniedGlobalKeys", "deletedDataRowGlobalKeys", "fetchedDataRows", "notFoundGlobalKeys"}
        existing_dr_to_gk       :   Dictinoary where {key=data_row_id : value=global_key} for data rows in use by global keys passed in. 
                                    If this = {}, then all global keys are free to use
    """
    query_keys = [str(x) for x in global_keys]
    # Create a query job to get data row IDs given global keys
    query_str_1 = """query get_datarow_with_global_key($global_keys:[ID!]!){dataRowsForGlobalKeys(where:{ids:$global_keys}){jobId}}"""
    query_str_2 = """query get_job_result($job_id:ID!){dataRowsForGlobalKeysResult(jobId:{id:$job_id}){data{
                    accessDeniedGlobalKeys\ndeletedDataRowGlobalKeys\nfetchedDataRows{id}\nnotFoundGlobalKeys}jobStatus}}"""        
    res = None
    while not res:
        query_job_id = client.execute(query_str_1, {"global_keys":global_keys})['dataRowsForGlobalKeys']['jobId']
        res = client.execute(query_str_2, {"job_id":query_job_id})['dataRowsForGlobalKeysResult']['data']       
    # If there are deleted data rows holding global keys, clear them and re-do the check
    if res["deletedDataRowGlobalKeys"]:
        client.clear_global_keys(res["deletedDataRowGlobalKeys"])   
        res = None
        while not res:
            query_job_id = client.execute(query_str_1, {"global_keys":global_keys})['dataRowsForGlobalKeys']['jobId']
            res = client.execute(query_str_2, {"job_id":query_job_id})['dataRowsForGlobalKeysResult']['data']  
    # Create a dictionary where {key=data_row_id : value=global_key}
    existing_dr_to_gk = {}
    for i in range(0, len(res["fetchedDataRows"])):
        data_row_id = res["fetchedDataRows"][i]["id"]
        global_key = global_keys[i]
        if data_row_id:
            existing_dr_to_gk[data_row_id] = global_key             
    return res, existing_dr_to_gk

def batch_create_data_rows(client:labelboxClient, upload_dict:dict, 
                           skip_duplicates:bool=True, divider:str="___", batch_size:int=20000, verbose:bool=False):
    """ Uploads data rows, skipping duplocate global keys or auto-generating new unique ones. Upload dict must be in the following format:
        {
            dataset_id : {
                global_key : {
                    "data_row" : {} -- This is your data row upload as a dictionary (must have keys "row_data", "global_key", at a minimum)
                }
                global_key : {
                    "data_row" : {}
                }
            },
            dataset_id : {
                global_key : {
                    "data_row" : {}
                }
                global_key : {
                    "data_row" : {}
                }
            }
        }    
    Args:
        client                                  :   Required (labelbox.client.Client) - Labelbox Client object
        upload_dict                             :   Required (dict) - Dictionary in the format outlined above
        skip_duplicates                         :   Optional (bool) - If True, will skip duplicate global_keys, otherwise will generate a unique global_key with a suffix "_1", "_2" and so on
        divider                                 :   Optional (str) - If skip_duplicates=False, uploader will auto-add a suffix to global keys to create unique ones, where new_global_key=old_global_key+divider+clone_counter
        batch_size                              :   Optional (int) - Upload batch size, 20,000 is recommended
        verbose                                 :   Optional (bool) - If True, prints information about code execution
    Returns:
        upload_errors                           :   Either a list Labelbox upload errors or an empty list if no errors
        updated_dict                            :   Updated dataset_to_global_key_to_upload_dict if global keys were removed or updated
    """
    updated_dict = {}
    for dataset_id in upload_dict.keys():
        dataset = client.get_dataset(dataset_id)
        global_key_to_upload_dict = upload_dict[dataset_id]
        global_keys_list = list(global_key_to_upload_dict.keys())
        payload, existing_data_row_to_global_key = check_global_keys(client, global_keys_list)
        if payload["deletedDataRowGlobalKeys"]:
            client.clear_global_keys(res["deletedDataRowGlobalKeys"])        
        payload, existing_data_row_to_global_key = check_global_keys(client, global_keys_list)
        loop_counter = 0
        while existing_data_row_to_global_key: # If existing data rows are using gloval keys
            if skip_duplicates: # Drop from global_key_to_upload_dict if we're skipping duplicates
                if verbose:
                    print(f"Warning: Global keys in this upload are in use by active data rows, skipping the upload of data rows affected") 
                for gk in existing_data_row_to_global_key.keys():
                    del global_key_to_upload_dict[gk]
                break
            else: # Create new suffix, replace in global_key_to_upload_dict, refresh existing_data_row_to_global_key
                loop_counter += 1 # Count the amount of attempts - this is used as a suffix                
                if verbose:
                    print(f"Warning: Global keys in this upload are in use by active data rows, attempting to add the following suffix to affected data rows: '{divider}{loop_counter}'")                   
                for gk in existing_data_row_to_global_key.values():
                    gk_root = gk if loop_counter == 1 else gk[:-(len(divider)+len(str(loop_counter)))] # Root global key, no suffix
                    new_gk = f"{gk_root}{divider}{loop_counter}" # New global key with suffix
                    upload_value = global_key_to_upload_dict[gk] # Grab data row upload
                    upload_value["data_row"]["global_key"] = new_gk # Replace the global key in our data row upload value
                    del global_key_to_upload_dict[gk] # Delete global key that's in use already
                    global_key_to_upload_dict[new_gk] = upload_value # Replace with new global key
                global_keys_list = list(global_key_to_upload_dict.keys()) # Make a new global key list
                payload, existing_data_row_to_global_key = check_global_keys(client, global_keys_list) # Refresh existing_data_row_to_global_key
        updated_dict[dataset_id] = global_key_to_upload_dict # Since we may have dropped/replaced some global keys, we will return a modified index            
        upload_list = [upload_value["data_row"] for upload_value in global_key_to_upload_dict.values()]
        if verbose:
            print(f'Beginning data row upload for dataset ID {dataset_id}: uploading {len(upload_list)} data rows')
        batch_number = 0
        for i in range(0,len(upload_list),batch_size):
            batch_number += 1 # Count uploads
            batch = upload_list[i:] if i + batch_size >= len(upload_list) else upload_list[i:i+batch_size] # Determine batch
            if verbose:
                print(f'Batch #{batch_number}: {len(batch)} data rows')
            task = dataset.create_data_rows(batch)
            task.wait_till_done()
            errors = task.errors
            if errors:
                if verbose: 
                    print(f'Error: Upload batch number {batch_number} unsuccessful')
                e = errors
                break
            else:
                if verbose: 
                    print(f'Success: Upload batch number {batch_number} successful')  
                e = []
    if verbose:
        print(f'Upload complete - all data rows uploaded')
    return e, updated_dict

def batch_upload_annotations(client:labelboxClient, project_id_to_upload_dict:dict, import_name:str=str(uuid.uuid4()), 
                             how:str="import", batch_size:int=10000, verbose=False):
    """ Batch imports labels given a batch size via MAL or LabelImport
    Args:
        client                      :   Required (labelbox.client.Client) - Labelbox Client object
        project_id_to_upload_dict   :   Required (dict) Dictionary where {key=project_id - value=annotation_upload_list} - annotations must be in ndjson format
        import_name                 :   Optional (str) - Name to give to import jobs - will have a batch number suffix
        how                         :   Optional (str) - Upload method - options are "mal" and "import" - defaults to "import"
        batch_size                  :   Optional (int) - Desired batch upload size - this size is determined by annotation counts, not by data row count
        verbose                     :   Optional (bool) - If True, prints information about code execution
    Returns: 
        A list of errors if there is one, True if upload failed, False if successful
    """
    # Determine the upload type
    if how.lower() == "mal":
        from labelbox import MALPredictionImport as upload_protocol
        if verbose:
            print(f"Uploading annotations as non-submitted pre-labels (MAL)")            
    elif how.lower() == "import":
        from labelbox import LabelImport as upload_protocol
        if verbose:
            print(f"Uploading annotations as submitted labels (Label Import)")
    else:
        return f"No annotation upload attempted - import method must be wither 'mal' or 'import' - received value {how}"
    batch_number = 0        
    # For each project
    if project_id_to_upload_dict:
        for project_id in project_id_to_upload_dict:
            annotations = project_id_to_upload_dict[project_id]
            # Create a dicationary where {key=data_row_id : value=list_of_annotations}
            data_row_id_to_upload_dict = {}
            for annotation in annotations:
                data_row_id = annotation['dataRow']['id']
                if data_row_id not in data_row_id_to_upload_dict:
                    data_row_id_to_upload_dict[data_row_id] = [annotation]
                else:
                    data_row_id_to_upload_dict[data_row_id].append(annotation)                   
            # Create ndjson batches at the data row level            
            data_row_list = list(data_row_id_to_upload_dict.keys())
            if verbose:
                print(f"Uploading {len(annotations)} annotations for {len(data_row_list)} data rows to project with ID {project_id}")             
            for i in range(0, len(data_row_list), batch_size):
                data_row_batch = data_row_list[i:] if i+batch_size >= len(data_row_list) else data_row_list[i:i+batch_size]
                upload = []
                for annotation in project_id_to_upload_dict[project_id]:
                    if annotation['dataRow']['id'] in data_row_batch:
                        upload.append(annotation)
                batch_number += 1
                import_request = upload_protocol.create_from_objects(client, project_id, f"{import_name}-{batch_number}", upload)
                errors = import_request.errors
                if errors:
                    if verbose:
                        print(f'Error: upload batch number {batch_number} unsuccessful')
                    return errors
                else:
                    if verbose:
                        print(f'Success: upload batch number {batch_number} complete')   
        return []
    else:
        return "No annotation upload attempted"


def batch_rows_to_project(client:labelboxClient, project_id_to_batch_dict:dict, priority:int=5, 
                          batch_name:str=str(uuid.uuid4()), batch_size:int=1000, verbose:bool=False):
    """ Takes a large amount of data row IDs and creates subsets of batches to send to a project
    Args:
        client                      :   Required (labelbox.client.Client) - Labelbox Client object
        project_id_to_batch_dict    :   Required (dict) Dictionary where {key=project_id : value=list_of_data_row_ids}
        priority                    :   Optinoal (int) - Between 1 and 5, what priority to give to data row batches sent to projects
        batch_name                  :   Optional (str) : Prefix to add to batch name - script generates batch number, which it adds to said prefix
        batch_size                  :   Optional (int) : Size of batches to send to project
        verbose                     :   Optional (bool) - If True, prints information about code execution        
    Returns:
        Empty list if successful, errors if unsuccessful
    """
    try:
        batch_number = 0
        for project_id in project_id_to_batch_dict:
            data_row_ids = project_id_to_batch_dict[project_id]
            if verbose:
                print(f"Sending {len(data_row_ids)} data rows to project with ID {project_id}")
            project = client.get_project(project_id)
            for i in range(0, len(data_row_ids), batch_size):
                batch_number += 1
                subset = data_row_ids[i:] if i+batch_size >= len(data_row_ids) else data_row_ids[i:i+batch_size]
                project.create_batch(name=f"{batch_name}-{batch_number}", data_rows=subset)
        if verbose:
            print(f"All data rows have been batched to the specified project(s)")
        return []  
    except Exception as e:
        print(f"Batching data rows to project unsuccessful: {e}")
        return e



from labelbox import Ontology as labelboxOntology

def get_ontology_schema_to_name_path(ontology, divider:str="///", invert:bool=False, detailed:bool=False):
    """ Recursively iterates through an ontology to create a dictionary where {key=schema_id : value=name_path}
    Where name_path = parent{divider}answer{divider}parent{divider}answer.... where divider="///"
    Args:
        ontology_normalized     :   Required (dict or labelbox.schema.ontology.Ontology) - Either Labelbox Ontology object or labelbox.schema.ontology.Ontology.normalized dictionary
        divider                 :   Optional (str) - String delineating the tool/classification/answer path for a given schema ID
        invert                  :   Optional (bool) - If True, will invert the dictionary to where {key=name_path : value=schema_id}
        detailed                :   Optional (bool) - If True, the dictionary will have the same key, but the values will be {"name" "color" "type" "kind" "parent_schema_ids"}
    Returns:
        Dictionary where {key=schema_id : value=name_path} - or the inverse - value can be more detailed if detailed=True
    """
    def map_layer(feature_dict:dict={}, node_layer:list= [], parent_name_path:str="", divider:str="///", invert:bool=False, detailed:bool=False, encoded_value:int=0):
        """ Recursive function that does the following for each node in a node_layer:
                1. Creates a name_path given the parent_name_path
                2. Adds schema_id : name_path to your working dictionary
                3. If there's another layer for a given node, recursively calls itself, passing it its own name key as it's childrens' parent_name_path
        Args:
            feature_dict        :   Required (dict) - Building dictionary of ontology information
            node_layer          :   Required (list) - A list of ontology classification, tool, or option dictionaries
            parent_name_path    :   Required (str) - A concatenated list of parent node names separated with `divider` value, creating a unique mapping key
            invert              :   Optional (bool) - If True, will invert the dictionary to where {key=name_path : value=schema_id}
            divider             :   Optional (str) - String delineating the tool/classification/answer path for a given schema ID
            detailed            :   Optional (bool) - If True, the dictionary will have the same key, but the values will be {"name" "color" "type" "kind" "parent_schema_ids"}
            encoded_value       :   Optional (int) - Counts the number of nodes in an ontology, creating an encoded numerical value for schemas
        Returns:
            feature_dict
        """
        if node_layer:
            for node in node_layer:
                encoded_value += 1
                if "tool" in node.keys():
                    node_name = node["name"]
                    next_layer = node["classifications"]
                    node_type = node["tool"]
                    node_type = "bbox" if node_type == "rectangle" else node_type
                    node_type = "mask" if node_type in ["superpixel", "raster-segmentation"] else node_type 
                    node_kind = "tool"   
                elif "instructions" in node.keys():
                    node_name = node["instructions"]
                    next_layer = node["options"]
                    node_kind = "classification"
                    node_type = node["type"]                        
                else:
                    node_type = "option"
                    node_name = node["label"]
                    next_layer = node.get("options", [])
                    node_kind = "branch_option" if next_layer else "leaf_option" 
                name_path = f"{parent_name_path}{divider}{node_name}" if parent_name_path else node_name
                dict_key = node['featureSchemaId'] if not invert else name_path
                if detailed:
                    if not invert:
                        dict_value = {"name":node_name,"type":node_type,"kind":node_kind,"encoded_value":encoded_value,"name_path":name_path}
                    else:
                        dict_value = {"name":node_name,"type":node_type,"kind":node_kind,"encoded_value":encoded_value,"schema_id":node['featureSchemaId']}
                else:
                    dict_value = name_path if not invert else node['featureSchemaId']
                feature_dict.update({dict_key : dict_value})
                if next_layer:
                    feature_dict, encoded_value = map_layer(feature_dict, next_layer, name_path, divider, invert=invert, detailed=detailed, encoded_value=encoded_value)
        return feature_dict, encoded_value
    if type(ontology) == labelboxOntology:
        ontology_normalized = ontology.normalized
    elif type(ontology) == dict:
        ontology_normalized = ontology
    else:
        raise TypeError(f"Input for ontology must be either a Lablbox ontology object or a dictionary representation of a Labelbox ontology - received input of type {ontology}") 
    if ontology_normalized["tools"]:
        working_dictionary, working_encoded_value = map_layer(feature_dict={}, node_layer=ontology_normalized["tools"], divider=divider, invert=invert, detailed=detailed)
    else:
        working_dictionary = {} 
        working_encoded_value = 0
    if ontology_normalized["classifications"]:
        working_dictionary, working_encoded_value = map_layer(feature_dict=working_dictionary, node_layer=ontology_normalized["classifications"], divider=divider, invert=invert, detailed=detailed, encoded_value=working_encoded_value)
    return working_dictionary



import numpy as np
from PIL import Image
from io import BytesIO
import requests
from labelbox.data import annotation_types as lb_types
from labelbox.data.serialization import NDJsonConverter

def mask_to_bytes(input:str, method:str="url", color=(255,255,255), output:str="png"):
    """ Given a mask input, returns a png bytearray of said mask a a dictionary
    Args:
        input     :   Required (str) - URL of a mask
        method    :   Required (str) - Either "url" or "array" - determines how you want the input treated
        color     :   Required (arr or int) - The color of your mask in your input value
        output    :   Required (str) - Either "array" or "png" - determines how you want the data returned
    Returns:
        Mask as a numpy array or as string png bytes
    """
    # Convert URL or array mask into a binary array
    if method not in ["url", "array"]:
        raise ValueError(f'Downloading bytes requires input method to be either a "url" or a "array" - received method {method}')
    if output not in ["array", "png"]:
        raise ValueError(f'Downloading bytes requires output method to be either a "png" or a "array" - received method {method}')     
    # Either download a mask URL or ensure the shape of your numpy array
    if method == "url":
        r = requests.get(input).content
        np_mask = np.array(Image.open(BytesIO(r)))[:,:,:3]
    else:
        if len(input.shape) == 3:
            np_mask = input
        elif len(input.shape) == 2:
            np_mask = [input, input, input]
        else:
            raise ValueError(f"Input segmentation mask arrays must either be 2D or 3D - shape of input mask: {input.shape}")
    if type(color) == int:
        np_color = (color, color, color)
    elif len(color) == 3:
        np_color = color
    else:
        raise ValueError(f"The specified color of your segmentation mask must either be a number (for 2D masks) or an RGB code (for 3D masks) - {color}")
    # Return either a numpy array or a png byte string
    if output == "array":
        return np_mask
    else:
        mask_label = lb_types.Label(
            data=lb_types.ImageData(uid=""),
            annotations=[
                lb_types.ObjectAnnotation(
                    name="", 
                    value=lb_types.Mask(
                        mask=lb_types.MaskData(arr=np_mask), 
                        color=color
                    )
                )
            ]
        )
        # Convert back into ndjson
        mask_png = list(NDJsonConverter.serialize([mask_label]))[0]["mask"]["png"]
        return mask_png



from labelbox import Client as labelboxClient
from labelbox.schema.data_row_metadata import DataRowMetadataKind
from labelbase.metadata import _refresh_metadata_ontology

def validate_columns(client:labelboxClient, table, get_columns_function, get_unique_values_function, 
                     divider:str="///", verbose:bool=False, extra_client=None):
    """ Given a table of columns with the right naming formats, does the following:
    
    1. Identifies a `row_data_col` name, `global_key_col` name, and an `external_id_col` name
    
    - `row_data_col` must be identified as the column with the name `row_data`
    - `global_key_col` defaults to `row_data_col` if not identified
    - `external_id_col` defaults to `global_key_col` if not identified
    
    2. Creates a metadata_index, attachment_index, and annotation_index - validates that the column names are in the right format:
    
    - Metadata columns must be `metadata{divider}metadata_type{divider}metadata_field_name`
        - metadata_type must be one of |"enum", "string", "datetime", "number"| (not case sensitive)
        
    - Attachment columns must be `attachment{divider}attachment_type{divider}column_name` (column_name is not relevant for attachment columns
        - metadata_type must be one of |"IMAGE", "VIDEO", "RAW_TEXT", "HTML", "TEXT_URL"| (not case sensitive)
        
    - Metadata columns must be `annotation{divider}annotation_type{divider}top_level_name`
        - annotation_type must be one of |"bbox", "polygon", "point", "mask", "line", "named-entity", "radio", "checklist", "text"| (not case sensitive)
        
    Args:
        client                      :   Required (labelbox.client.Client) - Labelbox Client object    
        table                       :   Required - Input user table    
        get_columns_function        :   Required (function) - Function that can get the column names from the table provided, returns list of strings
        get_unique_values_function  :   Required (function) - Function that grabs all unique values from a column, returns list of strings
        verbose                     :   Optional (bool) - If True, prints information about code execution
        extra_client                :   Optional - If relevant, the input get_columns_function / get_unique_values_function required other client object
    Returns:
        row_data_col                :   Raises an error if no `row_data` column is provided
        global_key_col              :   Defaults to row_data_col
        external_id_col             :   Defaults to global_key_col
        project_id_col              :   Returns "" if no `project_id` column is provided
        dataset_id_col              :   Returns "" if no `dataset_id` column is provided
        metadata_index              :   Dictonary where {key=metadata_field_name : value=metadata_type}
        attachment_index            :   Dictonary where {key=column_name : value=attachment_type}
        annotation_index            :   Dictonary where {key=column_name : value=annotation_type}
    """
    metadata_index = {}
    attachment_index = {}
    annotation_index = {}
    row_data_col = ""
    global_key_col = ""
    external_id_col = ""
    project_id_col = ""
    dataset_id_col = ""
    accepted_metadata_types = ["enum", "string", "datetime", "number"]
    accepted_attachment_types = ["IMAGE", "VIDEO", "RAW_TEXT", "HTML", "TEXT_URL"]
    accepted_annotation_types = ["bbox", "polygon", "point", "mask", "line", "named-entity", "radio", "checklist", "text"]
    column_names = get_columns_function(table=table, extra_client=extra_client)
    for column_name in column_names:
        if divider in column_name:
            input_type, column_type, header = column_name.split(divider)
            if input_type.lower() == "metadata":
                if column_type.lower() not in accepted_metadata_types:
                    raise ValueError(f"Invalid value in metadata column name {column_name} - must be `metadata{divider}` followed by one of the following: |{accepted_metadata_types}| followed by `{divider}metadata_field_name`")
                metadata_index[header] = column_type.lower()
            elif input_type.lower() == "attachment":
                if column_type.upper() not in accepted_attachment_types:
                    raise ValueError(f"Invalid value in attachment column name {column_name} - must be `attachment{divider}` followed by one of the following: |{accepted_attachment_types}| followed by `{divider}column_name`")
                attachment_index[column_name] = column_type.upper()
            elif input_type.lower() == "annotation":
                if column_type.lower() not in accepted_annotation_types:
                    raise ValueError(f"Invalid value in annotation column name {column_name} - must be `annotation{divider}` followed by one of the following: |{accepted_annotation_types}| followed by `{divider}top_level_feature_name`")
                annotation_index[column_name] = header
        else:
            if column_name == "row_data":
                row_data_col = "row_data"
            if column_name == "global_key":
                global_key_col = "global_key"
            if column_name == "external_id":
                external_id_col = "external_id"   
            if column_name == "project_id":
                project_id_col = "project_id"   
            if column_name == "dataset_id":
                dataset_id_col = "dataset_id" 
    if not row_data_col:
        raise ValueError(f"No `row_data` column found - please provide a column of row data URls with the colunn name `row_data`")
    global_key_col = global_key_col if global_key_col else row_data_col
    external_id_col = external_id_col if external_id_col else global_key_col
    lb_mdo, lb_metadata_names = _refresh_metadata_ontology(client)
    metadata_types = {
      "enum" : DataRowMetadataKind.enum, 
      "string" : DataRowMetadataKind.string, 
      "datetime" : DataRowMetadataKind.datetime, 
      "number" : DataRowMetadataKind.number
    }
    # If a metadata field name was passed in that doesn't exist, create it in Labelbox
    if metadata_index:
        for metadata_field_name in metadata_index.keys():
            metadata_string_type = metadata_index[metadata_field_name]
            if metadata_field_name not in lb_metadata_names:
                enum_options = get_unique_values_function(table=table, column_name=f"metadata{divider}{metadata_string_type}{divider}{metadata_field_name}", extra_client=extra_client) if metadata_string_type == "enum" else []
                if verbose:
                    print(f"Creating Labelbox metadata field with name {metadata_field_name} of type {metadata_string_type}")
                lb_mdo.create_schema(name=metadata_field_name, kind=metadata_types[metadata_string_type], options=enum_options)
    if "lb_integration_source" not in lb_metadata_names:
        lb_mdo.create_schema(name="lb_integration_source", kind=metadata_types["string"])
    return row_data_col, global_key_col, external_id_col, project_id_col, dataset_id_col, metadata_index, attachment_index, annotation_index



from labelbox import Client as labelboxClient
from labelbox.schema.data_row_metadata import DataRowMetadataKind
from datetime import datetime
from dateutil import parser
import pytz

def get_metadata_schema_to_type(client:labelboxClient, lb_mdo=False, invert:bool=False):
    """ Creates a dictionary where {key=metadata_schema_id: value=metadata_type} 
    - metadata_type either "string" "enum" "datetime" or "number"
    Args:
        client              :   Required (labelbox.client.Client) - Labelbox Client object    
        lb_mdo              :   Optional (labelbox.schema.data_row_metadata.DataRowMetadataOntology) - Labelbox metadata ontology
        divider             :   Optional (str) - String separating parent and enum option metadata values
        invert              :   Optional (bool) - If True, inverts the dictionary to be where {key=metadata_name_key: value=metadata_schema_id}
    Returns:
        Dictionary where {key=metadata_schema_id: value=metadata_type} - or the inverse
    """    
    metadata_schema_to_type = {}
    lb_mdo = client.get_data_row_metadata_ontology() if not lb_mdo else lb_mdo
    for field in lb_mdo._get_ontology():
        metadata_type = ""
        if "enum" in field["kind"].lower():
            metadata_type = "enum"
        if "string" in field["kind"].lower():
            metadata_type = "string"
        if "datetime" in field["kind"].lower():
            metadata_type = "datetime"
        if "number" in field["kind"].lower():
            metadata_type = "number"
        if metadata_type:
            metadata_schema_to_type[field["id"]] = metadata_type
    return_value = metadata_schema_to_type if not invert else {v:k for k,v in metadata_schema_to_type.items()}
    return return_value

def get_metadata_schema_to_name_key(client:labelboxClient, lb_mdo=False, divider="///", invert:bool=False):
    """ Creates a dictionary where {key=metadata_schema_id: value=metadata_name_key} 
    - name_key is name for all metadata fields, and for enum options, it is "parent_name{divider}child_name"
    Args:
        client              :   Required (labelbox.client.Client) - Labelbox Client object    
        lb_mdo              :   Optional (labelbox.schema.data_row_metadata.DataRowMetadataOntology) - Labelbox metadata ontology
        divider             :   Optional (str) - String separating parent and enum option metadata values
        invert              :   Optional (bool) - If True, inverts the dictionary to be where {key=metadata_name_key: value=metadata_schema_id}
    Returns:
        Dictionary where {key=metadata_schema_id: value=metadata_name_key} - or the inverse
    """
    lb_mdo = client.get_data_row_metadata_ontology() if not lb_mdo else lb_mdo
    lb_metadata_dict = lb_mdo.reserved_by_name
    lb_metadata_dict.update(lb_mdo.custom_by_name)
    metadata_schema_to_name_key = {}
    for metadata_field_name_key in lb_metadata_dict:
        if type(lb_metadata_dict[metadata_field_name_key]) == dict:
            metadata_schema_to_name_key[lb_metadata_dict[metadata_field_name_key][next(iter(lb_metadata_dict[metadata_field_name_key]))].parent] = str(metadata_field_name_key)
            for enum_option in lb_metadata_dict[metadata_field_name_key]:
                metadata_schema_to_name_key[lb_metadata_dict[metadata_field_name_key][enum_option].uid] = f"{str(metadata_field_name_key)}{str(divider)}{str(enum_option)}"
        else:
            metadata_schema_to_name_key[lb_metadata_dict[metadata_field_name_key].uid] = str(metadata_field_name_key)
    return_value = metadata_schema_to_name_key if not invert else {v:k for k,v in metadata_schema_to_name_key.items()}
    return return_value  

def _refresh_metadata_ontology(client:labelboxClient):
    """ Refreshes a Labelbox Metadata Ontology
    Args:
        client              :   Required (labelbox.client.Client) - Labelbox Client object    
    Returns:
        lb_mdo              :   labelbox.schema.data_row_metadata.DataRowMetadataOntology
        lb_metadata_names   :   List of metadata field names from a Labelbox metadata ontology
    """
    lb_mdo = client.get_data_row_metadata_ontology()
    lb_metadata_names = [field['name'] for field in lb_mdo._get_ontology()]
    return lb_mdo, lb_metadata_names

def sync_metadata_fields(client:labelboxClient, table, get_columns_function, add_column_function, get_unique_values_function, metadata_index:dict={}, verbose:bool=False, extra_client=None):
    """ Ensures Labelbox's Metadata Ontology and your input have all necessary metadata fields / columns given a metadata_index
    Args:
        client                      :   Required (labelbox.client.Client) - Labelbox Client object    
        table                       :   Required - Input user table
        get_columns_function        :   Required (function) - Function that can get the column names from the table provided, returns list of strings
        add_column_function         :   Required (function) - Function that can add an empty column to the table provided, returns table
        get_unique_values_function  :   Required (function) - Function that grabs all unique values from a column, returns list of strings
        metadata_index              :   Optional (dict) - Dictionary where {key=column_name : value=metadata_type} - metadata_type must be one of "enum", "string", "datetime" or "number"
        verbose                     :   Optional (bool) - If True, prints information about code execution
        extra_client                :   Optional - If relevant, the input functions' required other client object
    Returns:
        Updated table if successful, False if not
    """
    # Get your metadata ontology, grab all the metadata field names
    lb_mdo, lb_metadata_names = _refresh_metadata_ontology(client)
    # Convert your meatdata_index values from strings into labelbox.schema.data_row_metadata.DataRowMetadataKind types
    conversion = {"enum" : DataRowMetadataKind.enum, "string" : DataRowMetadataKind.string, "datetime" : DataRowMetadataKind.datetime, "number" : DataRowMetadataKind.number}
    # Check to make sure the value in your metadata index is one of the accepted values        
    _enforce_metadata_index(metadata_index, verbose)
    # If your table doesn't have columns for all your metadata_field_names, make columns for them
    if type(table) != bool:
        if metadata_index:
            column_names = get_columns_function(table, extra_client=extra_client)
            for metadata_field_name in metadata_index.keys():
                if metadata_field_name not in column_names:
                    table = add_column_function(table, column_name=metadata_field_name, default_value=None, extra_client=extra_client)
    # If Labelbox doesn't have metadata for all your metadata_field_names, make Labelbox metadata fields
    for metadata_field_name in metadata_index.keys():
        metadata_type = metadata_index[metadata_field_name]
        # Check to see if a metadata index input is a metadata field in Labelbox. If not, create the metadata field in Labelbox. 
        if metadata_field_name not in lb_metadata_names:
            enum_options = get_unique_values_function(table, metadata_field_name, extra_client=extra_client) if metadata_type == "enum" else []
            lb_mdo.create_schema(name=metadata_field_name, kind=conversion[metadata_type], options=enum_options)
            lb_mdo, lb_metadata_names = _refresh_metadata_ontology(client)
    if 'lb_integration_source' not in lb_metadata_names:
        lb_mdo.create_schema(name='lb_integration_source', kind=conversion["string"])
    return table  

def process_metadata_value(metadata_value, metadata_type:str, parent_name:str, metadata_name_key_to_schema:dict, divider:str="///"):
    """ Processes inbound values to ensure only valid values are added as metadata to Labelbox given the metadata type. Returns None if invalid or None
    Args:
        metadata_value              :   Required (any) - Value to-be-screeened and inserted as a proper metadata value to-be-uploaded to Labelbox
        metadata_type               :   Required (str) - Either "string", "datetime", "enum", or "number"
        parent_name                 :   Required (str) - Parent metadata field name
        metadata_name_key_to_schema :   Required (dict) - Dictionary where {key=metadata_field_name_key : value=metadata_schema_id}
        divider                     :   Required (str) - String delimiter for all name keys generated
    Returns:
        The proper data type given the metadata type for the input value. None if the value is invalud - should be skipped
    """
    if not metadata_value: # Catch empty values
        return_value = None
    if str(metadata_value) == "nan": # Catch NaN values
        return_value = None
    # By metadata type
    if metadata_type == "enum": # For enums, it must be a schema ID - if we can't match it, we have to skip it
        name_key = f"{parent_name}{divider}{str(metadata_value)}"
        if name_key in metadata_name_key_to_schema.keys():
            return_value = metadata_name_key_to_schema[name_key]
        else:
            return_value = None                  
    elif metadata_type == "number": # For numbers, it's ints as strings
        try:
            return_value = str(int(metadata_value))
        except:
            return_value = None                  
    elif metadata_type == "string": 
        return_value = str(metadata_value)
    else: # For datetime, it's an isoformat string
        if type(metadata_value) == str:
            return_value = parser.parse(metadata_value).astimezone(pytz.utc).replace(tzinfo=None)
        elif type(metadata_value) == datetime:
            return_value = metadata_value.astimezone(pytz.utc).replace(tzinfo=None)
        else:
            return_value = None       
    return return_value    



import uuid
import json
from labelbase.masks import mask_to_bytes

def get_leaf_paths(export_classifications:list, schema_to_name_path:dict, divider:str="///"):
    """ Given a flat list of labelox export classifications, constructs leaf name paths given a divider
    Args:
        export_classifications  :   Required (list) - List of classifications from label["Label"]["objects"][0]["classifications"] or label["Label"]["classificaitons"]
        schema_to_name_path     :   Required (dict) - Dictionary where {key=schema_id : value=feature_name_path} created from running:
                                            labelbase.get_ontology_schema_to_name_path(ontology, divider=divider, invert=False, detailed=False)
        divider                 :   Optional (str): String delimiter for name paths
    Returns:
        List of all leaf name paths 
    """
    def build_leaf_paths(root:dict, acc="", name_paths=[], divider="///"):
        for parent in root.keys():
            name_path = f"{acc}{divider}{parent}" if acc else f"{parent}"
            child = root[parent]
            if child:
                name_paths = build_leaf_paths(root=root[parent], acc=name_path, name_paths=name_paths)
            else:
                name_paths.append(name_path)
        return name_paths    
    name_paths = []
    for cla in export_classifications:
        if type(cla) == dict:
            if "answers" in cla.keys():
                for answer in cla["answers"]:
                    name_paths.append(schema_to_name_path[answer["schemaId"]])
            if "answer" in cla.keys():
                if type(cla["answer"]) == str:
                    name_paths.append(schema_to_name_path[cla["schemaId"]]+divider+cla["answer"])
                else:
                    name_paths.append(schema_to_name_path[cla["answer"]["schemaId"]]) 
        else:
            for c in cla:
                if "answers" in c.keys():
                    for answer in c["answers"]:
                        name_paths.append(schema_to_name_path[answer["schemaId"]])
                if "answer" in c.keys():
                    if type(c["answer"]) == str:
                        name_paths.append(schema_to_name_path[c["schemaId"]]+divider+c["answer"])
                    else:
                        name_paths.append(schema_to_name_path[c["answer"]["schemaId"]])                 
    root = {}
    for input_path in name_paths:
        parts = input_path.split(divider)
        current_node = root
        for part in parts:
            if part not in current_node:
                current_node[part] = {}
            current_node = current_node[part]    
    return build_leaf_paths(root)  

def flatten_label(label_dict:dict, ontology_index:dict, schema_to_name_path:dict, mask_method:str="url", divider:str="///"):
    """ For a label from project.export_labels(download=True), creates a flat dictionary where:
            { key = annotation_type + divider + annotation_name  :  value = [annotation_value, list_of_nested_name_paths]}
        Each accepted annotation type and the expected output annotation value is listed below:
            For tools:
                bbox            :   [[top, left, height, width], [nested_classification_name_paths], [top, left, height, width], [nested_classification_name_paths]]
                polygon         :   [[(x, y), (x, y),...(x, y)], [nested_classification_name_paths], [(x, y), (x, y),...(x, y)], [nested_classification_name_paths]]
                line            :   [[(x, y), (x, y),...(x, y)], [nested_classification_name_paths], [(x, y), (x, y),...(x, y)], [nested_classification_name_paths]]
                point           :   [[x, y], [nested_classification_name_paths], [x, y], [nested_classification_name_paths]]
                mask            :   [[URL, colorRGB], [nested_classification_name_paths], [URL, colorRGB], [nested_classification_name_paths]]
                                            OR
                                    [[array, colorRGB], [nested_classification_name_paths], [array, colorRGB], [nested_classification_name_paths]]
                                            OR
                                    [[png_bytes, "null"], [nested_classification_name_paths], [png_bytes, "null"], [nested_classification_name_paths]]                                    
                named-entity    :   [[start, end], [nested_classification_name_paths], [start, end], [nested_classification_name_paths]]
            For classifications:
                radio           :   [[answer_name_paths]]
                check           :   [[answer_name_paths]]
                text            :   [[answer_name_paths]] -- the last string in a text name path is the text value itself
    Args:    
        label_dict              :   Required (dict) - Dictionary representation of a label from project.export_labels(download=True)
        ontology_index          :   Required (dict) - Dictionary created from running:
                                            labelbase.ontology.get_ontology_schema_to_name_path(ontology, divider=divider, invert=True, detailed=True)
        schema_to_name_path     :   Required (dict) - Dictionary where {key=schema_id : value=feature_name_path} created from running:
                                            labelbase.ontology.get_ontology_schema_to_name_path(ontology, divider=divider, invert=False, detailed=False)
        mask_method             :   Optional (str) - Specifies your desired mask data format
                                        - "url" leaves masks as-is
                                        - "array" converts URLs to numpy arrays
                                        - "png" converts URLs to png byte strings                                               
        divider                 :   Optional (str) - String delimiter for name paths        
    Returns:        
        Dictionary with one key per annotation class in a given label in the specified format written above
    """
    flat_label = {}
    annotations = label_dict["Label"]
    objects = annotations["objects"]
    classifications = annotations["classifications"]
    if objects:
        for obj in objects:
            annotation_type = ontology_index[obj["title"]]["type"]
            annotation_type = "mask" if annotation_type == "raster-segmentation" else annotation_type
            annotation_type = "bbox" if annotation_type == "rectangle" else annotation_type
            column_name = f'{annotation_type}{divider}{obj["title"]}'           
            if column_name not in flat_label.keys():
                flat_label[column_name] = []
            if "bbox" in obj.keys():
                annotation_value = [obj["bbox"]["top"], obj["bbox"]["left"], obj["bbox"]["height"], obj["bbox"]["width"]]
            elif "polygon" in obj.keys():
                annotation_value = [[coord["x"], coord["y"]] for coord in obj["polygon"]]
            elif "line" in obj.keys():
                annotation_value = [[coord["x"], coord["y"]] for coord in obj["line"]]
            elif "point" in obj.keys():
                annotation_value = [obj["point"]["x"], obj["point"]["y"]]
            elif "location" in obj.keys():
                annotation_value = [obj["location"]["start"], obj["location"]["end"]]
            else:
                if mask_method == "url":
                    annotation_value = [obj["instanceURI"], [255,255,255]]
                elif mask_method == "array": 
                    array = mask_to_bytes(input=obj["instanceURI"], method="url", color=[255,255,255], output="array")
                    annotation_value = [array, [255,255,255]]
                else:
                    png = mask_to_bytes(input=obj["instanceURI"], method="url", color=[255,255,255], output="png")
                    annotation_value = [png, "null"]
            if "classifications" in obj.keys():
                nested_classification_name_paths = get_leaf_paths(
                    export_classifications=obj["classifications"], 
                    schema_to_name_path=schema_to_name_path,
                    divider=divider
                )
                return_paths = get_child_paths(first=obj["title"], name_paths=nested_classification_name_paths, divider=divider)
            else:
                return_paths = []
            flat_label[column_name].append([annotation_value, return_paths])
    if classifications:
        leaf_paths = get_leaf_paths(
            export_classifications=classifications, 
            schema_to_name_path=schema_to_name_path,
            divider=divider
        )
        classification_names = pull_first_name_from_paths(
            name_paths=leaf_paths, 
            divider=divider
        )
        for classification_name in classification_names:
            annotation_type = ontology_index[classification_name]["type"]
            child_paths = get_child_paths(first=classification_name, name_paths=leaf_paths, divider=divider)
            flat_label[f'{annotation_type}{divider}{classification_name}'] = [[name_path for name_path in child_paths]]
    return flat_label

def create_ndjsons(top_level_name:str, annotation_inputs:list, ontology_index:dict, mask_method:str="url", divider:str="///"):
    """ From an annotation in the expected format, creates a Labelbox NDJSON of that annotation -- note the data row ID is not added here
        Each accepted annotation type and the expected input annotation value is listed below:
            For tools:
                bbox            :   [[top, left, height, width], [nested_classification_name_paths], [top, left, height, width], [nested_classification_name_paths]]
                polygon         :   [[(x, y), (x, y),...(x, y)], [nested_classification_name_paths], [(x, y), (x, y),...(x, y)], [nested_classification_name_paths]]
                line            :   [[(x, y), (x, y),...(x, y)], [nested_classification_name_paths], [(x, y), (x, y),...(x, y)], [nested_classification_name_paths]]
                point           :   [[x, y], [nested_classification_name_paths], [x, y], [nested_classification_name_paths]]
                mask            :   [[URL, colorRGB], [nested_classification_name_paths], [URL, colorRGB], [nested_classification_name_paths]]
                                            OR
                                    [[array, colorRGB], [nested_classification_name_paths], [array, colorRGB], [nested_classification_name_paths]]
                                            OR
                                    [[png_bytes, None], [nested_classification_name_paths], [png_bytes, None], [nested_classification_name_paths]]                                    
                named-entity    :   [[start, end], [nested_classification_name_paths], [start, end], [nested_classification_name_paths]]
            For classifications:
                radio           :   [[answer_name_paths]]
                check           :   [[answer_name_paths]]
                text            :   [[answer_name_paths]] -- the last string in a text name path is the text value itself
    Args:
        data_row_id             :   Required (str) - Labelbox Data Row ID
        top_level_name          :   Required (str) - Name of the top-level tool or classification        
        annotation_inputs       :   Required (list) - List of annotation value lists where 1 list element must correspond to the following format:        
        annotation_type         :   Required (str) - Type of annotation in question - must be a string of one of the above options
        ontology_index          :   Required (dict) - Dictionary created from running:
                                            labelbase.ontology.get_ontology_schema_to_name_path(ontology, divider=divider, invert=True, detailed=True)
        mask_method             :   Optional (str) - Specifies your input mask data format
                                        - "url" means your mask is an accessible URL (must provide color)
                                        - "array" means your mask is a numpy array (must provide color)
                                        - "png" means your mask value is a png-string                                       
        divider                 :   Optional (str) - String delimiter for name paths        
    """
    if mask_method not in ["url", "png", "array"]:
        raise ValueError(f"Mask method must be either `url`, `png` or `array`")
    ndjsons = []
    if (type(annotation_inputs) == str) and (annotation_inputs!=""):
        annotation_inputs = json.loads(annotation_inputs.replace("'",'"').replace("None","null"))
    if type(annotation_inputs) == list:
        for annotation_input in annotation_inputs:
            ndjsons.append(ndjson_builder(
                top_level_name=top_level_name,
                annotation_input=annotation_input, 
                ontology_index=ontology_index, 
                mask_method=mask_method,
                divider=divider
            ))
    return ndjsons

def ndjson_builder(top_level_name:str, annotation_input:list, ontology_index:dict, mask_method:str="url", divider:str="///"):
    """ Returns an ndjson of an annotation given a list of values - the values needed differ depending on the annotation type
    Args:
        top_level_name          :   Required (str) - Name of the top-level tool or classification        
        annotation_input        :   Required (list) - List that corresponds to a single annotation for a label in the specified format
        ontology_index          :   Required (dict) - Dictionary created from running:
                                            labelbase.ontology.get_ontology_schema_to_name_path(ontology, divider=divider, invert=True, detailed=True)
        mask_method             :   Optional (str) - Specifies your input mask data format
                                        - "url" treats annotation input values as URLs uploads them directly
                                        - "array" converts the annotation input values into png bytes
                                        - "png" uploads png bytes directly
        divider                 :   Optional (str) - String delimiter for name paths        
    Returns
        NDJSON representation of an annotation
    """
    annotation_type = ontology_index[top_level_name]["type"]
    ndjson = {
        "uuid" : str(uuid.uuid4())
    }  
    # Catches tools
    if annotation_type in ["bbox", "polygon", "line", "point", "mask", "named-entity"]:
        ndjson["name"] = top_level_name
        if annotation_type == "bbox":
            ndjson[annotation_type] = {"top":annotation_input[0][0],"left":annotation_input[0][1],"height":annotation_input[0][2],"width":annotation_input[0][3]}
        elif annotation_type in ["polygon", "line"]:
            ndjson[annotation_type] = [{"x":xy_pair[0],"y":xy_pair[1]} for xy_pair in annotation_input[0]]     
        elif annotation_type == "point":
            ndjson[annotation_type] = {"x":annotation_input[0][0],"y":annotation_input[0][1]}
        elif annotation_type == "mask":
            if mask_method == "url": 
                ndjson[annotation_type] = {"instanceURI":annotation_input[0][0],"colorRGB":annotation_input[0][1]}
            elif mask_method == "array": # input masks as numpy arrays
                png = mask_to_bytes(input=annotation_input[0][0], method=mask_method, color=annotation_input[0][1], output="png")
                ndjson[annotation_type] = {"png":png}
            else: # Only one left is png
                ndjson[annotation_type] = {"png":annotation_input[0][0]}
        else: # Only one left is named-entity 
            ndjson["location"] = {"start" : annotation_input[0][0],"end":annotation_input[0][1]}
        if annotation_input[1]:
            ndjson["classifications"] = []
            classification_names = pull_first_name_from_paths(name_paths=annotation_input[1], divider=divider)
            for classification_name in classification_names:
                ndjson["classifications"].append(
                    classification_builder(
                        classification_path=classification_name,
                        answer_paths=get_child_paths(first=classification_name, name_paths=annotation_input[1], divider=divider),
                        ontology_index=ontology_index,
                        tool_name=top_level_name,
                        divider=divider
                    )
                )
    # Otherwise, the top level feature is a classification
    else:
        ndjson.update(
            classification_builder(
                classification_path=top_level_name, 
                answer_paths=annotation_input,
                ontology_index=ontology_index,
                divider=divider
            )
        )
    return ndjson    

def classification_builder(classification_path:str, answer_paths:list, ontology_index:dict, tool_name:str="", divider:str="///"):
    """ Given a classification path and all its child paths, constructs an ndjson.
        If the classification answer's paths have nested classifications, will recursuively call this function.
    """
    index_input = f"{tool_name}{divider}{classification_path}" if tool_name else classification_path
    c_type = ontology_index[index_input]["type"]
    c_name = classification_path.split(divider)[-1] if divider in classification_path else classification_path
    classification_ndjson = {
        "name" : c_name
    }
    if c_type == "radio":
        answer_name = pull_first_name_from_paths(name_paths=answer_paths, divider=divider)[0]
        classification_ndjson["answer"] = {
            "name" : answer_name
        }
        n_c_paths = get_child_paths(first=answer_name, name_paths=answer_paths, divider=divider)
        n_c_names = pull_first_name_from_paths(name_paths=n_c_paths, divider=divider)
        for n_c_name in n_c_names:
            if n_c_name:
                if "classifications" not in classification_ndjson["answer"].keys():
                    classification_ndjson["answer"]["classifications"] = []
                n_a_paths = get_child_paths(first=n_c_name, name_paths=n_c_paths, divider=divider)  
                classification_ndjson["answer"]["classifications"].append(
                    classification_builder(
                        classification_path=f"{classification_path}{divider}{answer_name}{divider}{n_c_name}",
                        answer_paths=n_a_paths,
                        ontology_index=ontology_index,
                        tool_name=tool_name,
                        divider=divider
                    )
                )
    elif c_type == "checklist":
        classification_ndjson["answers"] = []
        answer_names = pull_first_name_from_paths(name_paths=answer_paths, divider=divider)
        for answer_name in answer_names:
            answer_ndjson = {
                "name" : answer_name
            }
            n_c_paths = get_child_paths(first=answer_name, name_paths=answer_paths, divider=divider)
            n_c_names = pull_first_name_from_paths(name_paths=n_c_paths, divider=divider)
            for n_c_name in n_c_names:
                if n_c_name:
                    if "classifications" not in answer_ndjson.keys():
                        answer_ndjson["classifications"] = []
                    n_a_paths = get_child_paths(first=n_c_name, name_paths=n_c_paths, divider=divider) 
                    answer_ndjson["classifications"].append(
                        classification_builder(
                            classification_path=f"{classification_path}{divider}{answer_name}{divider}{n_c_name}",
                            answer_paths=n_a_paths,
                            ontology_index=ontology_index,
                            tool_name=tool_name,
                            divider=divider                            
                        )
                    )
            classification_ndjson["answers"].append(answer_ndjson)
    else:
        classification_ndjson["answer"] = answer_paths[0]
    return classification_ndjson

def pull_first_name_from_paths(name_paths:list, divider:str="///"):
    """ Pulls the first name from every name path in a list a divider-delimited name paths
    Args:    
        name_paths              :   Required (list) - List of name paths
        divider                 :   Optional (str) - String delimiter for all name keys generated for parent/child schemas     
    Returns:
        List of unique first names from a given name path
    """    
    firsts = []
    for name_path in name_paths:
        firsts.append(str(name_path.split(divider)[0]))
    return list(set(firsts))

def get_child_paths(first, name_paths, divider:str="///"):
    """ From a list of name paths, grabs paths starting with the `first` string and removes the `first` name from the name path
    Args
        first                   :   Required (str) - The parent feature name you want to find paths for
        name_paths              :   Required (list) - List of name paths
        divider                 :   Optional (str) - String delimiter for all name keys generated for parent/child schemas             
    Returns
        List of children name paths
    """
    child_paths = []
    for path in name_paths:
        if path.startswith(first):
            child_path = ""
            for name in path.split(divider)[1:]:
                child_path += str(name)+str(divider)
            child_path = child_path[:-len(divider)] 
            child_paths.append(child_path)
    return child_paths      



from labelbox import Client as labelboxClient
from labelbox import Project as labelboxProject
from labelbase.ontology import get_ontology_schema_to_name_path
from labelbase.metadata import get_metadata_schema_to_name_key, get_metadata_schema_to_type
from labelbase.annotate import flatten_label

def export_and_flatten_labels(
    client:labelboxClient, project, include_metadata:bool=True, include_performance:bool=True, 
    include_agreement:bool=False, verbose:bool=False, mask_method:str="png", divider="///"):
    """ Exports and flattens labels from a Labelbox Project
    Args:
        client:                 :   Required (labelbox.Client) - Labelbox Client object
        project                 :   Required (str / lablebox.Project) - Labelbox Project ID or lablebox.Project object to export labels from
        include_metadata        :   Optional (bool) - If included, exports metadata fields
        include_performance     :   Optional (bool) - If included, exports labeling performance
        include_agreement       :   Optional (bool) - If included, exports consensus scores       
        verbose                 :   Optional (bool) - If True, prints details about code execution; if False, prints minimal information
        mask_method             :   Optional (str) - Specifies your input mask data format
                                        - "url" leaves masks as-is
                                        - "array" converts URLs to numpy arrays
                                        - "png" converts URLs to png byte strings        
        divider                 :   Optional (str) - String delimiter for schema name keys and suffix added to duplocate global keys 
    Returns:
        List of dictionaries where { key = column_name : value = row_value }
    """
    if mask_method not in ["url", "png", "array"]:
        raise ValueError(f"Please specify the mask_method you want to download your segmentation masks in - must be either 'url' 'png' or 'array'")
    project = project if type(project) == labelboxProject else client.get_project(project)
    if verbose:
        print(f"Exporting labels from Labelbox for project with ID {project.uid}")
    export = project.export_labels(download=True)
    if verbose:
        print(f"Export complete: {len(export)} labels exported")   
    if include_metadata:
        data_row_ids = list(set([label['DataRow ID'] for label in export]))
        if verbose:
            print(f"Exporting metadata from Labelbox for {len(data_row_ids)} data row IDs")
        mdo = client.get_data_row_metadata_ontology()
        metadata_export = mdo.bulk_export(data_row_ids=data_row_ids)
        metadata_export_index = {x.data_row_id : x for x in metadata_export}
        metadata_schema_to_type = get_metadata_schema_to_type(client=client, lb_mdo=mdo, invert=False)
        metadata_schema_to_name_key = get_metadata_schema_to_name_key(client=client, lb_mdo=mdo, invert=False, divider=divider)
        if verbose:
            print(f"Metadata export complete")
    ontology_index = get_ontology_schema_to_name_path(project.ontology(), invert=True, divider=divider, detailed=True)
    schema_to_name_path = get_ontology_schema_to_name_path(project.ontology(), invert=False, divider=divider, detailed=False)   
    flattened_labels = [] 
    if verbose:
        print(f"Flattening labels...")
    for label in export:
        if not label['Skipped']:
            flat_label = {
                "global_key" : label["Global Key"],
                "row_data" : label["Labeled Data"],
                "data_row_id" : label["DataRow ID"],
                "label_id" : label["ID"],
                "external_id" : label["External ID"]
            }
            res = flatten_label(label_dict=label, ontology_index=ontology_index, schema_to_name_path=schema_to_name_path, mask_method=mask_method, divider=divider)            
            for key, val in res.items():
                flat_label[f"annotation{divider}{str(key)}"] = val
            if include_agreement:
                flat_label["consensus_score"] = label["Agreement"]
            if include_performance:
                flat_label["created_by"] = label["Created By"]
                flat_label["seconds_to_create"] = label["Seconds to Create"]
                flat_label["seconds_to_review"] = label["Seconds to Review"]
                flat_label["seconds_to_label"] = label["Seconds to Label"]
            if include_metadata:
                data_row_metadata = metadata_export_index[label["DataRow ID"]].fields
                for metadata in data_row_metadata:
                    metadata_type = metadata_schema_to_type[metadata.schema_id]
                    if metadata.value in metadata_schema_to_name_key.keys():
                        name_path = metadata_schema_to_name_key[metadata.value].split(divider)
                        field_name = name_path[0]
                        metadata_value = name_path[1]
                    else:
                        field_name = metadata.name
                        metadata_value = metadata.value
                    if field_name != "lb_integration_source":
                        flat_label[f'metadata{divider}{metadata_type}{divider}{field_name}'] = metadata_value
            flattened_labels.append(flat_label)
    if verbose:
        print(f"Labels flattened")            
    return flattened_labels 
  
  



from labelbox import Project as labelboxProject
from labelbase.ontology import get_ontology_schema_to_name_path
import copy
import datetime
from google.api_core import retry
import requests
from PIL import Image 
import numpy as np
from io import BytesIO
from tqdm import tqdm
from concurrent.futures import ThreadPoolExecutor, as_completed
from shapely.geometry import Polygon
import cv2

def _to_coco_bbox_converter(data_row_id:str, annotation:dict, category_id:str):
    """ Given a label dictionary and a bounding box annotation from said label, will return the coco-converted bounding box annotation dictionary
    Args:
        data_row_id (str)               :     Labelbox Data Row ID for this label
        annotation (dict)               :     Annotation dictionary from label['Label']['objects'], which comes from project.export_labels()
        category_id (str)               :     Desired category_id for the coco_annotation
    Returns:
        An annotation dictionary in the COCO format
    """
    coco_annotation = {
        "image_id": data_row_id,
        "bbox": [annotation['bbox']['left'], annotation['bbox']['top'], annotation['bbox']['width'], annotation['bbox']['height']],
        "category_id": category_id,
        "id": annotation['featureId']
    }
    return coco_annotation

def _to_coco_line_converter(data_row_id:str, annotation:dict, category_id:str):
    """ Given a label dictionary and a line annotation from said label, will return the coco-converted line annotation dictionary
    Args:
        data_row_id (str)               :     Labelbox Data Row ID for this label
        annotation (dict)               :     Annotation dictionary from label['Label']['objects'], which comes from project.export_labels()
        category_id (str)               :     Desired category_id for the coco_annotation
    Returns:
        An annotation dictionary in the COCO format
    """
    line = annotation['line']
    coco_line = []
    num_line_keypoints = 0
    for coordinates in line:
        coco_line.append(coordinates['x'])
        coco_line.append(coordinates['y'])
        coco_line.append(2)
        num_line_keypoints += 1
    coco_annotation = {
        "image_id": data_row_id,
        "keypoints": coco_line,
        "num_keypoints": num_line_keypoints,
        "category_id" : category_id,
        "id": annotation['featureId']
    }
    return coco_annotation, num_line_keypoints

def _to_coco_point_converter(data_row_id:str, annotation:dict, category_id:str):
    """ Given a label dictionary and a point annotation from said label, will return the coco-converted point annotation dictionary
    Args:
        data_row_id (str)               :     Labelbox Data Row ID for this label
        annotation (dict)               :     Annotation dictionary from label['Label']['objects'], which comes from project.export_labels()
        category_id (str)               :     Desired category_id for the coco_annotation
    Returns:
        An annotation dictionary in the COCO format
    """  
    coco_annotation = {
        "image_id": data_row_id,
        "keypoints": [annotation['point']['x'], annotation['point']['y'], 2],
        "num_keypoints": 1,
        "category_id" : category_id,
        "id": annotation['featureId']
    }
    return coco_annotation  

def _to_coco_polygon_converter(data_row_id:str, annotation:dict, category_id:str):
    """Given a label dictionary and a point annotation from said label, will return the coco-converted polygon annotation dictionary
    Args:
        data_row_id (str)               :     Labelbox Data Row ID for this label
        annotation (dict)               :     Annotation dictionary from label['Label']['objects'], which comes from project.export_labels()
        category_id (str)               :     Desired category_id for the coco_annotation
    Returns:
        An annotation dictionary in the COCO format
    """  
    poly_points = [[coord['x'], coord['y']] for coord in annotation['polygon']]
    bounds = Polygon(poly_points).bounds
    coco_annotation = {
        "image_id" : data_row_id,
        "segmentation" : [[item for sublist in poly_points for item in sublist]],
        "bbox" : [bounds[0], bounds[1], bounds[2]-bounds[0], bounds[3]-bounds[1]],
        "area" : Polygon(poly_points).area,
        "id": annotation['featureId'],
        "iscrowd" : 0,
        "category_id" : category_id
    }  
    return coco_annotation
  
@retry.Retry(predicate=retry.if_exception_type(Exception), deadline=120.)
def download_mask(url:str):
    """Incorporates retry logic into the download of a mask / polygon Instance URI
    Args:
        annotation (dict)       :     A dictionary pertaining to a label exported from Labelbox
    Returns:
        A numPy array of said mask
    """ 
    return requests.get(url).content 

def _to_coco_mask_converter(data_row_id:str, annotation:dict, category_id:str):
    """Given a label dictionary and a mask annotation from said label, will return the coco-converted segmentation mask annotation dictionary
    Args:
        data_row_id (str)               :     Labelbox Data Row ID for this label
        annotation (dict)               :     Annotation dictionary from label['Label']['objects'], which comes from project.export_labels()
        category_id (str)               :     Desired category_id for the coco_annotation
    Returns:
        An annotation dictionary in the COCO format
    """  
    mask_data = download_mask(annotation['instanceURI'])
    binary_mask_arr = np.array(Image.open(BytesIO(mask_data)))[:,:,:1]
    contours = cv2.findContours(binary_mask_arr, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)[0]
    coords = contours[0]
    poly_points = np.array([[coords[i][0][0], coords[i][0][1]] for i in range(0, len(coords))])
    bounds = Polygon(poly_points).bounds
    coco_annotation = {
        "image_id" : data_row_id,
        "segmentation" : [[item for sublist in poly_points for item in sublist]],
        "bbox" : [bounds[0], bounds[1], bounds[2]-bounds[0], bounds[3]-bounds[1]],
        "area" : Polygon(poly_points).area,
        "id": annotation['featureId'],
        "iscrowd" : 0,
        "category_id" : category_id
    }  
    return coco_annotation

def _to_coco_annotation_converter(data_row_id:str, annotation:dict, ontology_schema_to_name_path:dict):
    """ Wrapper to triage and multithread the coco annotation conversion - if nested classes exist, the category_id will be the first radio/checklist classification answer available
    Args:
        data_row_id (str)                   :     Labelbox Data Row ID for this label
        annotation (dict)                   :     Annotation dictionary from label['Label']['objects'], which comes from project.export_labels()
        ontology_schema_to_name_path (dict) :     A dictionary where {key=featureSchemaId : value = {"encoded_value"} which corresponds to category_id
    Returns:
        A dictionary corresponding to te coco annotation syntax - the category ID used will be the top-level tool 
    """
    max_line_keypoints = 0
    category_id = ontology_schema_to_name_path[annotation['schemaId']]['encoded_value']
    if "classifications" in annotation.keys():
        if annotation['classifications']:
            for classification in annotation['classifications']:
                if 'answer' in classification.keys():
                    if type(classification['answer']) == dict:
                        category_id = ontology_schema_to_name_path[classification['schemaId']]['encoded_value']
                        break
                else:
                    category_id = ontology_schema_to_name_path[classification['answers'][0]['schemaId']]['encoded_value']
                    break
    if "bbox" in annotation.keys():
        coco_annotation = _to_coco_bbox_converter(data_row_id, annotation, category_id)
    elif "line" in annotation.keys():
        coco_annotation, max_line_keypoints = _to_coco_line_converter(data_row_id, annotation, category_id)
    elif "point" in annotation.keys():
        coco_annotation = _to_coco_point_converter(data_row_id, annotation, category_id)
    elif "polygon" in annotation.keys():
        coco_annotation = _to_coco_polygon_converter(data_row_id, annotation, category_id)
    else: 
        coco_annotation = _to_coco_mask_converter(data_row_id, annotation, category_id)     
    return coco_annotation, max_line_keypoints

def export_labels(project:labelboxProject, verbose:bool=True, divider:str="///"):
    """ Given a project and a list of labels, will create the COCO export json
    Args:
        project     :   Required (labelbox.schema.project.Project) - Labelbox Project object
        verbose     :   Optional (bool) - If True, prints information about code execution
        divider     :   Optional (str) - String delineating the tool/classification/answer path for a given schema ID
    Returns:
        Dicationary with 'info', 'licenses', 'images', 'annotations', and 'annotations' keys corresponding to a COCO dataset format
    """
    if verbose:
        print(f'Exporting labels from project ID {project.uid}')
    labels_list = project.export_labels(download=True)
    if verbose:
        print(f'Export complete - {len(labels_list)} labels to convert')    
    info = {
        'description' : project.name, 'url' : f'https://app.labelbox.com/projects/{project.uid}/overview', 'version' : "1.0", 
        'year' : datetime.datetime.now().year, 'contributor' : project.created_by().email, 'date_created' : datetime.datetime.now().strftime('%Y/%m/%d'),
    }
    licenses = [{"url" : "N/A","id" : 1,"name" : "N/A"}]
    # Create a dictionary where {key=data_row_id : value=data_row}
    data_rows = {}
    if verbose:
        print(f'Exporting Data Rows from Project...')
    if project.datasets():
        for dataset in project.datasets():
            for data_row in dataset.export_data_rows():
                data_rows.update({data_row.uid : data_row})  
    else:
        for batch in project.batches():
            for data_row in batch.export_data_rows():
                data_rows.update({data_row.uid : data_row})  
    if verbose:                        
        print(f'Export complete. {len(data_rows)} Data Rows Exported')
        print(f'Converting Data Rows into a COCO Dataset...')
    images = []
    data_row_check = [] # This is a check for projects where one data row has multiple labels (consensus, benchmark)
    if verbose:
        for label in tqdm(labels_list):
            data_row = data_rows[label['DataRow ID']]
            if label['DataRow ID'] not in data_row_check:
                data_row_check.append(label['DataRow ID'])
                images.append({
                    "license" : 1, "file_name" : data_row.external_id, "height" : data_row.media_attributes['height'],
                    "width" : data_row.media_attributes['width'], "date_captured" : data_row.created_at.strftime('%Y-%m-%d %H:%M:%S'),
                    "id" : data_row.uid, "coco_url": data_row.row_data
                })
    else:
        for label in labels_list:
            data_row = data_rows[label['DataRow ID']]
            if label['DataRow ID'] not in data_row_check:
                data_row_check.append(label['DataRow ID'])
                images.append({
                    "license" : 1, "file_name" : data_row.external_id, "height" : data_row.media_attributes['height'],
                    "width" : data_row.media_attributes['width'], "date_captured" : data_row.created_at.strftime('%Y-%m-%d %H:%M:%S'),
                    "id" : data_row.uid, "coco_url": data_row.row_data
                })  
    if verbose:                     
        print(f'Data Rows Converted into a COCO Dataset.')  
    annotations = []
    if verbose:                     
        print(f'Converting Annotations into the COCO Format...')
    ontology_schema_to_name_path = get_ontology_schema_to_name_path(project.ontology().normalized, detailed=True) 
    global_max_keypoints = 0
    futures = []
    if verbose:
        with ThreadPoolExecutor() as exc:
            for label in tqdm(labels_list):
                for annotation in label['Label']['objects']:
                    futures.append(exc.submit(_to_coco_annotation_converter, label['DataRow ID'], annotation, ontology_schema_to_name_path))
            for f in tqdm(as_completed(futures)):
                res = f.result()
                if res[1] > global_max_keypoints:
                    global_max_keypoints = copy.deepcopy(res[1])
                annotations.append(res[0])    
    else:
        with ThreadPoolExecutor() as exc:
            for label in labels_list:
                for annotation in label['Label']['objects']:
                    futures.append(exc.submit(_to_coco_annotation_converter, label['DataRow ID'], annotation, ontology_schema_to_name_path))
            for f in as_completed(futures):
                res = f.result()
                if res[1] > global_max_keypoints:
                    global_max_keypoints = copy.deepcopy(res[1])
                annotations.append(res[0])          
    if verbose:                     
        print(f'Annotation Conversion Complete. Converted {len(annotations)} annotations into the COCO Format.')                        
        print(f'Converting the Ontology into the COCO Dataset Format...') 
    categories = []        
    for featureSchemaId in ontology_schema_to_name_path:
        if ontology_schema_to_name_path[featureSchemaId]["type"] == "line": 
            keypoints = []
            skeleton = []
            for i in range(0, global_max_keypoints): 
                keypoints.append(str("line_")+str(i+1))
                skeleton.append([i, i+1])
            categories.append({
                "supercategory" : ontology_schema_to_name_path[featureSchemaId]['name'],
                "id" : ontology_schema_to_name_path[featureSchemaId]["encoded_value"],
                "name" : ontology_schema_to_name_path[featureSchemaId]['name'],
                "keypoints" : keypoints, "skeleton" : skeleton,
            })
        elif ontology_schema_to_name_path[featureSchemaId]["type"] == "point": 
            categories.append({
                "supercategory" : ontology_schema_to_name_path[featureSchemaId]['name'],
                "id" : ontology_schema_to_name_path[featureSchemaId]["encoded_value"],
                "name" : ontology_schema_to_name_path[featureSchemaId]['name'],
                "keypoints" : ['point'], "skeleton" : [0, 0],
            })        
        elif ontology_schema_to_name_path[featureSchemaId]['kind'] == 'tool':
            categories.append({
                "supercategory" : ontology_schema_to_name_path[featureSchemaId]['name'],
                "id" : ontology_schema_to_name_path[featureSchemaId]["encoded_value"],
                "name" : ontology_schema_to_name_path[featureSchemaId]['name']
            })     
        elif len(ontology_schema_to_name_path[featureSchemaId]['name_path'].split(divider)) == 2:
            categories.append({
                "supercategory" : ontology_schema_to_name_path[featureSchemaId]['name_path'].split(divider)[0],
                "id" : ontology_schema_to_name_path[featureSchemaId]["encoded_value"],
                "name" : ontology_schema_to_name_path[featureSchemaId]['name']
            })
    if verbose:            
        print(f'Ontology Conversion Complete')       
    if verbose:            
        print(f'COCO Conversion Complete')    
    return {"info" : info, "licenses" : licenses, "images" : images, "annotations" : annotations, "categories" : categories}      



name = "labelbase"
__version__ = "0.0.51"

from labelbase import converters
from labelbase import uploader
from labelbase import downloader
from labelbase import metadata
from labelbase import ontology
from labelbase import dataset
from labelbase import annotate
from labelbase import connector
from labelbase import masks



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
      name='labelbase',
      version='0.0.83',
      author='Labelbox',
      author_email='raphael@labelbox.com',
      description='Labelbox Helper Library',      
      packages=setuptools.find_packages(),
      url="https://labelbox.com",
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires=["labelbox", "packaging"],
      keywords=["labelbox", "labelbase"],
      extras_require={'dev': ['pylint']}
)



from labelbox import Client as labelboxClient
from labelbox.schema.dataset import Dataset as labelboxDataset

def get_or_create_dataset(client:labelboxClient, name:str, integration:str="DEFAULT", verbose:bool=False):
    """ Gets or creates a Labelbox dataset given a dataset name and a deleagted access integration name
    Args:
        name                :   Required (str) - Desired dataset name
        integration         :   Optional (str) - Existing Labelbox delegated access setting for new dataset
        verbose             :   Optional (bool) - If True, prints information about code execution
    Returns:
        labelbox.schema.dataset.Dataset object
    """
    try: 
        dataset = next(client.get_datasets(where=(labelboxDataset.name==name)))
        if verbose:
            print(f'Got existing dataset with ID {dataset.uid}')
    except:
        for iam_integration in client.get_organization().get_iam_integrations(): 
            if iam_integration.name == integration: # If the names match, reassign the iam_integration input value
                integration = iam_integration
                if verbose:
                    print(f'Creating a Labelbox dataset with name "{name}" and delegated access integration name {integration.name}')
                break
        # If none match, use the default setting
        if (type(integration)==str) and (verbose==True):
            print(f'Creating a Labelbox dataset with name "{name}" and the default delegated access integration setting')
        # Create the Labelbox dataset 
        dataset = client.create_dataset(name=name, iam_integration=integration)             
        if verbose:
            print(f'Created a new dataset with ID {dataset.uid}') 
    return dataset
